# kate: syntax Python;
"""
Quality trimming.
"""

import sys

def quality_trim_index(str qualities, int cutoff, int base=33):
	"""
	Find the position at which to trim a low-quality end from a nucleotide sequence.

	Qualities are assumed to be ASCII-encoded as chr(qual + base).

	The algorithm is the same as the one used by BWA within the function
	'bwa_trim_read':
	- Subtract the cutoff value from all qualities.
	- Compute partial sums from all indices to the end of the sequence.
	- Trim sequence at the index at which the sum is minimal.
	"""
	cdef int s = 0
	cdef int max_qual = 0
	cdef int max_i = len(qualities)
	cdef int i
	cdef int q
	cdef bytes quals = qualities.encode()
	cdef char* cq = quals
	for i in reversed(range(len(quals))):
		q = cq[i] - base
		s += cutoff - q
		if s < 0:
			break
		if s > max_qual:
			max_qual = s
			max_i = i
	return max_i


def expected_errors(str qualities, int base=33):
	cdef int i, q
	cdef bytes quals = qualities.encode()
	cdef char* cq = quals
	cdef double e = 0.0
	for i in range(len(qualities)):
		q = cq[i] - base
		e += 10 ** (-q / 10)
	return e



"""
Open compressed files transparently.
"""
from __future__ import print_function, division, absolute_import
__author__ = 'Marcel Martin'

import gzip
import sys
import io
from subprocess import Popen, PIPE
from ..compat import PY3, basestring

try:
	import bz2
except ImportError:
	bz2 = None

if sys.version_info < (2, 7):
	buffered_reader = lambda x: x
	buffered_writer = lambda x: x
else:
	buffered_reader = io.BufferedReader
	buffered_writer = io.BufferedWriter


class GzipWriter:
	def __init__(self, path):
		self.outfile = open(path, 'w')
		try:
			self.process = Popen(['gzip'], stdin=PIPE, stdout=self.outfile)
		except IOError as e:
			self.outfile.close()
			raise

	def write(self, arg):
		self.process.stdin.write(arg)

	def close(self):
		self.process.stdin.close()
		c = self.process.wait()
		if c != 0:
			raise IOError("Output gzip process terminated with exit code {0}".format(c))


def xopen(filename, mode='r'):
	"""
	Replacement for the "open" function that can also open files that have
	been compressed with gzip or bzip2. If the filename is '-', standard
	output (mode 'w') or input (mode 'r') is returned. If the filename ends
	with .gz, the file is opened with a pipe to the gzip program. If that
	does not work, then gzip.open() is used (the gzip module is slower than
	the pipe to the gzip program). If the filename ends with .bz2, it's
	opened as a bz2.BZ2File. Otherwise, the regular open() is used.

	mode can be: 'rt', 'rb', 'wt', or 'wb'
	Instead of 'rt' and 'wt', 'r' and 'w' can be used as abbreviations.

	In Python 2, the 't' and 'b' characters are ignored.
	"""
	if mode == 'r':
		mode = 'rt'
	elif mode == 'w':
		mode = 'wt'
	if mode not in ('rt', 'rb', 'wt', 'wb'):
		raise ValueError("mode '{0}' not supported".format(mode))
	if not PY3:
		mode = mode[0]
	if not isinstance(filename, basestring):
		raise ValueError("the filename must be a string")

	# standard input and standard output handling
	if filename == '-':
		if not PY3:
			return sys.stdin if 'r' in mode else sys.stdout
		return dict(
			rt=sys.stdin,
			wt=sys.stdout,
			rb=sys.stdin.buffer,
			wb=sys.stdout.buffer)[mode]

	if filename.endswith('.bz2'):
		if bz2 is None:
			raise ImportError("Cannot open bz2 files: The bz2 module is not available")
		if PY3:
			if 't' in mode:
				return io.TextIOWrapper(bz2.BZ2File(filename, mode[0]))
			else:
				return bz2.BZ2File(filename, mode)
		else:
			return bz2.BZ2File(filename, mode)

	elif filename.endswith('.gz'):
		if PY3:
			if 't' in mode:
				return io.TextIOWrapper(gzip.open(filename, mode[0]))
			else:
				if 'r' in mode:
					return io.BufferedReader(gzip.open(filename, mode))
				else:
					return io.BufferedWriter(gzip.open(filename, mode))
		else:
			# rb/rt are equivalent in Py2
			if 'r' in mode:
				try:
					return Popen(['gzip', '-cd', filename], stdout=PIPE).stdout
				except IOError:
					# gzip not installed
					return buffered_reader(gzip.open(filename, mode))
			else:
				try:
					return GzipWriter(filename)
				except IOError:
					return buffered_writer(gzip.open(filename, mode))
	else:
		return open(filename, mode)



#!/usr/bin/env python3
"""
Generate random sequences in FASTA format
"""
import sys
from random import choice, randint

from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("--minimum-length", "-m", type=int, default=20)
	parser.add_argument("--maximum-length", "-M", type=int, default=50)
	parser.add_argument("n", type=int, help="Number of sequences to generate")

	args = parser.parse_args()
	ALPHABET = 'ACGT'
	for i in range(args.n):
		l = randint(args.minimum_length, args.maximum_length)
		seq = ''.join(choice('ACGT') for _ in range(l))
		print(">seq{0}\n{1}".format(i+1, seq))


if __name__ == '__main__':
	main()



__version__ = '0.4.5'

from .args import HelpfulArgumentParser
from .io.fasta import (
	SequenceReader, FastaReader, FastqReader, FastaWriter, FastqWriter,
	IndexedFasta, guess_quality_base )
from .io.gtf import GtfReader
from .cigar import Cigar
#from .align import multialign, consensus



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import sys


class HelpfulArgumentParser(ArgumentParser):
	"""An ArgumentParser that prints full help on errors."""

	def __init__(self, *args, **kwargs):
		if 'formatter_class' not in kwargs:
			kwargs['formatter_class'] = RawDescriptionHelpFormatter
		super().__init__(*args, **kwargs)

	def error(self, message):
		self.print_help(sys.stderr)
		args = {'prog': self.prog, 'message': message}
		self.exit(2, '%(prog)s: error: %(message)s\n' % args)



#!/usr/bin/env python3
"""
Print a comparison matrix for sequences from a FASTA or text file.
All pairs of sequences are compared using the given comparison measure.
By default, this is edit distance, the minimum number of substitutions,
insertions, and deletions to transform one sequence into the other.

For the resulting matrix, a header row and column is printed;
this can be disabled by the options below.

The given list of sequences can be automatically extended by their
reverse complements, appending a suffix like _RC to each sequence's name.
"""

__author__ = "Sven Rahmann"

import sys
from sqt import HelpfulArgumentParser
from sqt.io.fasta import SequenceReader
from sqt.align import edit_distance
from sqt.dna import reverse_complement as revcomp


def get_argument_parser():
    parser = HelpfulArgumentParser(description=__doc__)
    parser.add_argument("file",
        help="file with sequences to compare")
    parser.add_argument("-c", "--compareby", default="editdistance",
        choices=("none", "identity", "editdistance"),
        help="comparison function [default: editdistance]")
    parser.add_argument("-s", "--show", default="matrix",
        choices=("none", "matrix", "min", "max"),
        help="output type (matrix or nearest neighbors)")
    parser.add_argument("-r", "--revcomp", nargs="?", const="_RC", metavar="SUFFIX",
        help="compare reverse complements, too; append _RC or argument to each name")
    parser.add_argument("-f", "--format", choices=("text","fasta","fastq"),
        help="explicitly specify file format if not FASTA")
    parser.add_argument("-R", "--norowheaders", action="store_true",
        help="do not print initial names in each row")
    parser.add_argument("-C", "--nocolumnheaders", action="store_true",
        help="do not print first row with column headers")
    parser.add_argument("-H", "--noheaders", action="store_true",
        help="both --norowheaders and --nocolumnheaders")
    parser.add_argument("-F", "--featurename", default="feature", metavar="NAME",
        help="use the given argument instead of 'feature' for features")
    parser.add_argument("-S", "--separator", default="\t", metavar="SEPARATOR",
        help="field separator [default: TAB]")
    return parser


comparison_function = dict(
    none = lambda s,t: None,
    identity = lambda s,t: int(s==t),
    editdistance = edit_distance
)


def process_file(fname, fformat):
    """return list of tuples (name, seq) from file <fname>"""
    if fformat=="text":
        with open(fname, "rt") as f:
            return [("seq_"+str(i), line.strip()) for line in f.readlines()]
    # now it's fasta or fastq
    seqs = []
    for record in SequenceReader(fname):
        name, _, _ = record.name.partition(" ")
        seq = record.sequence.upper()
        seqs.append((name,seq))
    return seqs


def show_matrix(sequences, args):
    featurename = args.featurename
    separator = args.separator
    compare = comparison_function[args.compareby]
    if not args.nocolumnheaders:
        clist = [] if args.norowheaders else [featurename]
        clist.extend([s[0] for s in sequences])
        print(separator.join(clist))
    for si in sequences:
        clist = [] if args.norowheaders else [si[0]]
        for sj in sequences:
            c = compare(si[1],sj[1])
            clist.append(str(c))
        print(separator.join(clist))


def show_neighbors(sequences, args):
    assert args.show in ("max", "min")
    domax = (args.show == "max")
    separator = args.separator
    compare = comparison_function[args.compareby]
    for i,si in enumerate(sequences):
        clist = [] if args.norowheaders else [si[0]]
        best = None;  bestlist = []
        for j,sj in enumerate(sequences):
            if i==j: continue
            c = compare(si[1],sj[1])
            if domax:
                if best is None or c > best:
                    best = c;  bestlist=[sj[0]]
                elif c == best:
                    bestlist.append(sj[0])
            else:
                if best is None or c < best:
                    best = c;  bestlist=[sj[0]]
                elif c == best:
                    bestlist.append(sj[0])
        clist.append(str(best))
        print(separator.join(clist+bestlist))


show_functions = dict(
    none = lambda sequence, args: None,
    matrix = show_matrix,
    min = show_neighbors,
    max = show_neighbors,
)


def main():
    parser = get_argument_parser()
    args = parser.parse_args()
    sequences = process_file(args.file, args.format)
    if args.revcomp:
        rcsfx = args.revcomp
        rcs = [(name+rcsfx, revcomp(seq)) for (name,seq) in sequences]
        sequences.extend(rcs)
    fshow = show_functions[args.show]
    fshow(sequences, args)


if __name__ == '__main__':
    main()



"""
Minimal Py2/Py3 compatibility library.
"""
import sys
PY3 = sys.version > '3'


if PY3:
	maketrans = bytes.maketrans
	basestring = str
	zip = zip

	def bytes_to_str(s):
		return s.decode('ascii')

	def str_to_bytes(s):
		return s.encode('ascii')

	def force_str(s):
		if isinstance(s, (bytes, bytearray)):
			return s.decode('ascii')
		else:
			return s

else:
	def bytes_to_str(s):
		return s

	def str_to_bytes(s):
		return s

	def force_str(s):
		return s

	from string import maketrans
	basestring = basestring
	from itertools import izip as zip



"""
Quality trimming and filtering.
"""
import sys
if sys.version_info[0] >= 3:
	xrange = range


def quality_trim_index(qualities, cutoff, base=33):
	"""
	Find the position at which to trim a low-quality end from a nucleotide sequence.

	Qualities are assumed to be ASCII-encoded as chr(qual + base).

	The algorithm is the same as the one used by BWA within the function
	'bwa_trim_read':
	- Subtract the cutoff value from all qualities.
	- Compute partial sums from all indices to the end of the sequence.
	- Trim sequence at the index at which the sum is minimal.
	"""
	s = 0
	max_qual = 0
	max_i = len(qualities)
	for i in range(len(qualities)-1, -1, -1):
		q = ord(qualities[i]) - base
		s += cutoff - q
		if s < 0:
			break
		if s > max_qual:
			max_qual = s
			max_i = i
	return max_i


def expected_errors(qualities, base=33):
	"""
	Return expected number of errors.

	qualities -- ASCII-encoded qualities (chr(qual + base)).
	"""
	return sum(10 ** (-(ord(q) - base) / 10) for q in qualities)


try:
	from ._qualtrim import quality_trim_index, expected_errors
except:
	pass



import subprocess
from collections import Counter
from .io.fasta import FastaReader
from io import StringIO

try:
	from sqt._helpers import globalalign
except ImportError:
	import warnings
	warnings.warn("C module for speedups unavailable", ImportWarning)


class GlobalAlignment:
	"""A global alignment between two strings"""

	def __init__(self, s, t, semiglobal=False):
		if type(s) is str:
			s = s.encode('ascii')
		if type(t) is str:
			t = t.encode('ascii')
		flags = 15 if semiglobal else 0  # TODO this constant shouldn't be here
		(row1, row2, start1, stop1, start2, stop2, errors) = globalalign(s, t, flags)
		assert semiglobal or start1 == start2 == 0
		assert semiglobal or stop1 == len(s)
		assert semiglobal or stop2 == len(t)
		assert len(row1) == len(row2)
		self.row1 = row1.decode('ascii')
		self.row2 = row2.decode('ascii')
		self.start1 = start1
		self.stop1 = stop1
		self.start2 = start2
		self.stop2 = stop2
		self.errors = errors
		# Overlap start and stop in alignment coordinates.
		# row1[overlap_start:overlap_stop] is the overlapping alignment.
		self.overlap_start = max(start1, start2)
		self.overlap_stop = len(row1) - max(len(s) - stop1, len(t) - stop2)

	def _replace_gaps(self, row, gap_char):
		start = self.overlap_start
		stop = self.overlap_stop
		front = row[:start].replace('\0', ' ')
		middle = row[start:stop].replace('\0', gap_char)
		end = row[stop:].replace('\0', ' ')
		return front + middle + end

	def print(self, width=100, gap_char='-'):
		"""multi-line output and parameters, therefore not called __str__"""
		row1 = self._replace_gaps(self.row1, gap_char)
		row2 = self._replace_gaps(self.row2, gap_char)
		for i in range(0, len(row1), width):
			top = row1[i:i+width]
			bottom = row2[i:i+width]
			m = []
			for c, d in zip(top, bottom):
				if c == d:
					m.append('|')
				elif c == ' ' or d == ' ':
					m.append(' ')
				else:
					m.append('X')
			middle = ''.join(m)
			if i > 0:
				print()
			print(top)
			print(middle)
			print(bottom)

	def __str__(self):
		return 'GA(row1={}, row2={}, errors={})'.format(
			self.row1, self.row2, self.errors)


def edit_distance(s, t):
	"""
	Return the edit distance between the strings s and t.
	The edit distance is the sum of the numbers of insertions, deletions,
	and mismatches that is minimally necessary to transform one string
	into the other.
	"""
	m = len(s) # index: i
	n = len(t) # index: j

	# dynamic programming "table" (just a single column)
	# note that using an array('h', ...) here is not faster
	costs = list(range(m+1))

	# calculate alignment (using unit costs)
	prev = 0
	for j in range(1, n+1):
		prev = costs[0]
		costs[0] += 1
		for i in range(1, m+1):
			c = min(
				prev + int(s[i-1] != t[j-1]),
				costs[i] + 1,
				costs[i-1] + 1)
			prev = costs[i]
			costs[i] = c
	return costs[-1]


try:
	from sqt._helpers import edit_distance
except:
	pass





import os
import re
import subprocess


def available_cpu_count():
	"""
	Number of available virtual or physical CPUs on this system.

	Adapted from http://stackoverflow.com/a/1006301/715090
	"""
	# cpuset may restrict the number of available processors
	try:
		m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
						open('/proc/self/status').read())
		if m:
			res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
			if res > 0:
				return res
	except IOError:
		pass

	try:
		import multiprocessing
		return multiprocessing.cpu_count()
	except (ImportError, NotImplementedError):
		pass
	return 1


def multialign(sequences, program='mafft', threads=available_cpu_count()):
	"""
	Wrapper for multiple sequence alignment tools. Currently supported are
	* ClustalO (http://www.clustal.org/omega/),
	* MAFFT (http://mafft.cbrc.jp/alignment/software/)
	* MUSCLE (http://www.drive5.com/muscle/)

	A package using the libclustalo library directly also exists:
	https://github.com/benchling/clustalo-python/blob/master/clustalo.c
	It is not for Python 3.

	sequences -- a dictionary mapping names to sequences.
		Use an OrderedDict if order matters.

	program -- must be 'clustalo', 'mafft', 'muscle', 'muscle-medium',
		'muscle-fast'. The latter calls MUSCLE with parameters that make it run
		faster (but less accurate). 'muscle-medium' is in between muscle and
		muscle-fast.

	threads -- number of threads to use for those programs that support it.
		By default, set to the number of processors.
	"""
	if program == 'mafft':
		args = ['mafft', '--quiet', '--thread', str(threads), '-']
	elif program == 'clustalo':
		args = ['clustalo', '--threads='+str(threads), '--infile=-']
	elif program == 'muscle':
		args = ['muscle', '-quiet', '-in', '-', '-out', '-']
	elif program == 'muscle-fast':
		args = ['muscle', '-quiet', '-maxiters', '1', '-diags', '-in', '-', '-out', '-']
	elif program == 'muscle-medium':
		args = ['muscle', '-quiet', '-maxiters', '2', '-diags', '-in', '-', '-out', '-']
	else:
		raise ValueError('program {!r} not supported'.format(program))

	fasta_data = ''.join('>{}\n{}\n'.format(name, seq) for name, seq in sequences.items())
	result = subprocess.check_output(args, input=fasta_data, universal_newlines=True)
	aligned = list(FastaReader(StringIO(result)))

	return { record.name: record.sequence.upper() for record in aligned }


def consensus(aligned, threshold=0.7, ambiguous='N', keep_gaps=False):
	"""
	Compute a consensus from multialign() output.

	Idea taken from BioPython’s SummaryInfo.dumb_consensus function.

	keep_gaps -- whether the returned sequence contains gaps (-)
	"""
	result = []
	n = len(aligned)
	ambiguous = 'N'
	for chars in zip(*aligned.values()):
		char, freq = Counter(chars).most_common(1)[0]
		if freq / n >= threshold:
			if keep_gaps or char != '-':
				result.append(char)
		else:
			result.append(ambiguous)
	return ''.join(result)



#!/usr/bin/env python3
"""
Print lots of statistics about one or more FASTA files.

TODO
- computation of contig N50 is relatively slow
"""
import sys
import os
import subprocess
from collections import Counter
from sqt.io.fasta import FastaReader, IndexedFasta
from sqt.dna import n_intervals, intervals_complement
from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def byte_frequencies(s):
	return Counter(s)


try:
	from sqt._helpers import byte_frequencies
except:
	pass


def n50(lengths, genome_size=None):
	"""
	Return N50 or NG50 value given a list of lengths.

	If the genome_size is not given, it is set to sum(lengths). The resulting
	value is the N50. If it is given, the resulting value is the NG50.

	If genome_size is greater than 2 * sum(lengths), None is returned.
	"""
	if genome_size is None:
		genome_size = sum(lengths)
	lengths = sorted(lengths, reverse=True)
	running_total = 0
	for length in lengths:
		running_total += length
		if running_total >= genome_size * 0.5:
			return length
	return None


def print_statistics(lengths, contig_lengths, names, genome_size=None, character_frequencies=None):
	n = len(lengths)
	total = sum(lengths)
	assert len(lengths) == len(names)
	print('No. of sequences: {:11,}'.format(len(lengths)))
	if not lengths:
		return
	min_length, max_length = min(lengths), max(lengths)
	min_name = names[lengths.index(min_length)]
	max_name = names[lengths.index(max_length)]
	lengths.sort()
	print('Total length:   {:13,}'.format(sum(lengths)))
	print('Minimum length: {:13,} in entry "{}"'.format(min_length, min_name))
	print('Maximum length: {:13,} in entry "{}"'.format(max_length, max_name))
	print('Average length: {:16.2f}'.format(sum(lengths) / len(lengths)))
	print('Median length:  {:13,}'.format(lengths[n // 2]))
	print('Scaffold N50:   {:13,}'.format(n50(lengths)))
	if genome_size:
		print('Scaffold NG50:  {:13,}'.format(n50(lengths, genome_size=genome_size)))

	if contig_lengths:
		lengths = contig_lengths
		min_length, max_length = min(lengths), max(lengths)
		lengths.sort()
		print()
		print('Number of contigs:     {:13,}'.format(len(lengths)))
		print('Total contig length:   {:13,}'.format(sum(lengths)))
		print('Minimum contig length: {:13,}'.format(min_length))
		print('Maximum contig length: {:13,}'.format(max_length))
		print('Average contig length: {:16.2f}'.format(sum(lengths) / len(lengths)))
		print('Median contig length:  {:13,}'.format(lengths[n // 2]))
		print('Contig N50:            {:13,}'.format(n50(lengths)))

	if character_frequencies:
		print()
		print("Character distribution (<char> <count> <percentage>):")
		assert total == sum(character_frequencies.values())
		acgt = 0
		gc = 0
		for upper, lower in (b'Aa', b'Cc', b'Gg', b'Tt'):
			freq = character_frequencies[upper] + character_frequencies[lower]
			if upper in b'GC':
				gc += freq
			print(chr(upper), '    {:14,} {:6.1%}'.format(freq, freq / total))
			acgt += freq
		other = sum(character_frequencies.values()) - acgt
		print('other {:14,} {:6.2%}'.format(other, other / total))
		print('ACGT  {:14,} {:6.2%}'.format(acgt, acgt / total))
		print('GC    {:14,} {:6.2%} (of ACGT)'.format(gc, gc / (total - other)))


def get_lengths(path):
	"""
	Return a list of all sequence lengths of a FASTA file.

	The information is read from the FASTA index file (.fai). If it does
	not exist, an attempt is made to create it.
	"""
	if not os.path.exists(path + '.fai'):
		print('FASTA index file (.fai) does not exist, running "samtools faidx" ...', file=sys.stderr)
		try:
			subprocess.check_output(['samtools', 'faidx', path])
		except (subprocess.CalledProcessError, OSError) as e:
			print('samtools faidx failed:', e)
			raise  # TODO or just do not use the index
	lengths = []
	names = []
	with IndexedFasta(path) as f:
		for index_entry in f.index.values():
			lengths.append(index_entry.length)
			names.append(index_entry.name)
	return lengths, names


def scaffold_stats(path):
	"""Return a tuple (lengths, names, character_frequencies)."""
	lengths = []
	names = []
	counter = Counter()
	with FastaReader(path, 'rb') as reader:
		for record in reader:
			seq = record.sequence #.upper()
			lengths.append(len(seq))
			names.append(record.name)
			counter += byte_frequencies(seq)
	return lengths, names, counter


def filter_short_intervals(intervals, minimum_length):
	for start, stop in intervals:
		if stop - start >= minimum_length:
			yield (start, stop)


def stats(path, tolerable_gapsize):
	"""
	Determine scaffold lengths, contig lengths and character frequencies.

	Return a tuple (scaffold_lengths, contig_lengths, names, character_frequencies).
	"""
	scaffold_lengths = []
	contig_lengths = []
	names = []
	counter = Counter()
	with FastaReader(path, 'rb') as reader:
		for record in reader:
			seq = record.sequence #.upper()
			scaffold_lengths.append(len(seq))
			names.append(record.name)
			counter += byte_frequencies(seq)
			intervals = intervals_complement(
				filter_short_intervals(n_intervals(seq, ord(b'N')), tolerable_gapsize), len(seq))
			for start, stop in intervals:
				contig_lengths.append(stop - start)

	return scaffold_lengths, contig_lengths, names, counter


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	add = parser.add_argument
	add('--detailed', '-d', default=False, action='store_true',
		help='Print information about the sequences themselves, '
			'such as the character distribution and contig N50.')
	add('--genome-size', '-g', type=int, default=None,
		help='Estimated genome size. If given, also NG50 in addition to N50 is computed.')
	add('--tolerable-gapsize', '-t', type=int, default=10,
		help='A stretch of at most this many "N"s is not counted as a gap '
		'separating contigs.')
	add('fasta', nargs='+', metavar='FASTA',
		help='Input FASTA [[TODO or FASTQ file(s)]] (may be gzipped).')
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()

	overall_frequencies = Counter()
	overall_lengths = []  # for scaffold lengths
	overall_names = []
	if not args.detailed:
		character_frequencies = None
		contig_lengths = None
	for path in args.fasta:
		print("## File:", path)
		if args.detailed:
			scaffold_lengths, contig_lengths, names, character_frequencies = \
				stats(path, args.tolerable_gapsize)
			overall_frequencies += character_frequencies
		else:
			scaffold_lengths, names = get_lengths(path)
		print_statistics(scaffold_lengths, contig_lengths, names, args.genome_size, character_frequencies)
		overall_lengths.extend(scaffold_lengths)
		overall_names.extend(names)
	if len(args.fasta) > 1:
		print("## Summary of", len(args.fasta), "files")
		print_statistics(overall_lengths, None, overall_names, args.genome_size, overall_frequencies if args.detailed else None)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Check whether two FASTQ files contain paired-end reads.
The corresponding reads in FASTQ1 and FASTQ2 must have
the same names. A read name suffix of "/1" in FASTQ1 and
a suffix of "/2" in FASTQ is ignored.

The read name is only the part before the first space character.

FASTQ1 and FASTQ2 may be gzipped.

The exit code is nonzero on failure, zero otherwise.
"""
import sys
from itertools import zip_longest
from sqt.io.fasta import FastqReader
from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def check(fastq1, fastq2, limit=None):
	"""
	Raise ValueError if given FASTQ files seem to be paired incorrectly.
	"""
	with FastqReader(fastq1) as fastq1:
		with FastqReader(fastq2) as fastq2:
			n = 1
			for record1, record2 in zip_longest(fastq1, fastq2):
				if record1 is None:
					raise ValueError("File 2 is longer than file 1")
				if record2 is None:
					raise ValueError("File 1 is longer than file 2")
				name1 = record1.name.split(' ', 1)[0]
				if name1.endswith("/1"):
					name1 = name1[:-2]
				name2 = record2.name.split(' ', 1)[0]
				if name2.endswith("/2"):
					name2 = name2[:-2]
				if name1 != name2:
					raise ValueError("Incorrect names at record no. {}: '{}' != '{}'".format(n, name1, name2))
				if n == limit:
					return
				n += 1


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("-n", "--limit", type=int, default=1000,
		help="Only check the first N records. Set to zero to check the entire file (default: %(default)s)")
	parser.add_argument("--quiet", "-q", action='store_true', default=False,
		help="Don't print anything, just set the exit code.")
	parser.add_argument('fastq1', metavar='FASTQ1',
		help='File with first read of paired-end reads')
	parser.add_argument('fastq2', metavar='FASTQ2',
		help='File with second read of paired-end reads')
	args = parser.parse_args()

	if not args.quiet:
		def message(*args, **kwargs):
			print(*args, **kwargs)
	else:
		def message(*args, **kwargs):
			pass

	try:
		check(args.fastq1, args.fastq2, args.limit)
	except ValueError as e:
		message(e)
		sys.exit(1)
	message("OK")


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
from collections import namedtuple
from struct import calcsize, unpack, unpack_from
import gzip
from itertools import chain
import sys

from ..cigar import Cigar

Reference = namedtuple('Reference', ['name', 'length'])

class FormatError:
	pass

SEQ_TRANS = bytes.maketrans(bytes(range(16)), b'=ACMGRSVTWYHKDBN')
CIGAR_OPS = 'MIDNSHP=X'
QUAL_TRANS = bytes.maketrans(bytes(range(100)), bytes(range(33, 133)))

def quality_to_ascii(qualities):
	return qualities.translate(QUAL_TRANS).decode('ascii')


TAG_TYPES = {
	'A': 'c',  # character
	'c': 'b',  # signed 8-bit integer
	'C': 'B',  # unsigned 8-bit integer
	's': 'h',  # signed 16-bit integer
	'S': 'H',  # unsigned 16-bit integer
	'i': 'i',  # signed 32-bit integer
	'I': 'I',  # unsigned 32-bit integer
	'f': 'f',  # 32-bit floating point number
	#'Z': zero-terminated string
	#'H': byte array in hex format
	#'B': integer or numeric array
}

class BamAlignment:
	def __init__(self, data, references):
		self.references = references
		align = '<iiIIiiii'
		(
			self.reference_id,
			self.position,
			bin_mq_nl,
			flag_nc,
			l_seq,
			self.next_reference_id,
			self.next_position,
			self.insert_length
		) = unpack_from(align, data)
		#bin = bin_mq_nl >> 16

		self.mapping_quality = (bin_mq_nl & 0xFF00) >> 8
		self.flags = flag_nc >> 16
		n = calcsize(align)

		# Query name
		query_name_length = bin_mq_nl & 0xFF
		self.query_name = data[n: n + query_name_length - 1].decode('ascii') # NULL-terminated
		if self.query_name == '*':
			self.query_name = None
		offset = n + query_name_length

		# CIGAR
		n_cigar_op = flag_nc & 0xFFFF
		cigar_codes = unpack_from('<{}I'.format(n_cigar_op), data, offset=offset)
		if not cigar_codes:
			self.cigar = None
		else:
			self.cigar = Cigar((code & 0xF, code >> 4) for code in cigar_codes)
		offset += 4 * n_cigar_op

		# Sequence
		encoded_seq = data[offset:offset + (l_seq + 1) // 2]
		sequence = bytes(chain(*[(v>>4, v&0xF) for v in encoded_seq]))
		sequence = sequence.translate(SEQ_TRANS).decode('ascii')
		if l_seq & 1 == 1:  # if odd
			sequence = sequence[:l_seq]
		self.sequence = sequence
		offset += (l_seq + 1) // 2

		# Qualities
		qualities = data[offset:offset + l_seq]
		if qualities[0] == 255:
			qualities = None
		self.qualities = qualities
		offset += l_seq

		self.tags = self._parse_tags(data, offset)

	def _parse_tags(self, data, offset):
		tags = []
		while offset < len(data):
			assert offset < len(data) - 4
			tag_name = data[offset:offset+2].decode('ascii')
			tag_type = chr(data[offset+2])
			assert tag_type in 'AcCsSiIfZHB'
			if tag_type == 'Z':
				i = data.index(0, offset+3)
				tag_value = data[offset+3:i].decode('ascii')
				offset = i + 1
			elif tag_type == 'H':
				raise NotImplementedError("Tag type H not implemented")
			elif tag_type == 'B':
				raise NotImplementedError("Tag type B not implemented")
			else:
				tag_value = unpack_from(TAG_TYPES[tag_type], data, offset + 3)[0]
				offset += 3 + calcsize(TAG_TYPES[tag_type])
			tags.append((tag_name, tag_type, tag_value))
		return tags

	@property
	def qualities_string(self):
		if self.qualities is None:
			return None
		return self.qualities.translate(QUAL_TRANS).decode('ascii')

	def _id_to_reference_name(self, id):
		assert id < len(self.references)
		if id >= 0:
			return self.references[id].name
		else:
			return None

	@property
	def reference_name(self):
		return self._id_to_reference_name(self.reference_id)

	@property
	def next_reference_name(self):
		return self._id_to_reference_name(self.next_reference_id)


		ref_name = self.references[ref_id].name if ref_id >= 0 else '*'
		if next_ref_id == ref_id:
			next_ref_name = '='
		else:
			next_ref_name = self.references[next_ref_id].name if next_ref_id >= 0 else '*'
		# note: pos and next_pos can have value -1

	def __str__(self):
		"""Return SAM-formatted representation of this record"""
		refname = self.reference_name
		if self.next_reference_id == self.reference_id and self.reference_id >= 0:
			nextrefname = '='
		else:
			nextrefname = self.next_reference_name

		tags = '\t'.join(':'.join((name, type, str(value))) for name, type, value in self.tags)
		fields = [
			self.query_name,
			self.flags,
			refname if refname is not None else '*',
			self.position + 1,
			self.mapping_quality,
			self.cigar if self.cigar else '*',
			nextrefname if nextrefname else '*',
			self.next_position + 1,
			self.insert_length,
			self.sequence,
			quality_to_ascii(self.qualities),
			tags
		]
		return '\t'.join(str(f) for f in fields)


class BamReader:
	MAGIC = b'BAM\1'

	def __init__(self, file):
		"""
		open the file, read the header
		"""
		file = gzip.GzipFile(file, 'rb')
		data = file.read(4)
		if data != self.MAGIC:
			raise FormatError("magic bytes 'BAM\\1' not found, is this a BAM file?")
		# header in SAM text format
		header_length = unpack('<i', file.read(4))[0]
		self.header = file.read(header_length).decode('ascii')
		# BAM header with reference sequence names and lengths
		n = unpack('<i', file.read(4))[0]
		refs = []
		for i in range(n):
			ref_name_length = unpack('<i', file.read(4))[0]
			ref_name = file.read(ref_name_length)[:-1].decode('ascii') # NULL-terminated
			ref_length = unpack('<i', file.read(4))[0]
			refs.append(Reference(ref_name, ref_length))
		self.references = refs
		self.file = file

	def __iter__(self):
		while True:
			data = self.file.read(4)
			if len(data) < 4:
				break
			block_size = unpack('<i', data)[0]
			data = self.file.read(block_size)
			yield BamAlignment(data, self.references)


def main():
	reader = BamReader(sys.argv[1])
	print(reader.header, end='')
	for alignment in reader:
		try:
			print(alignment)
		except BrokenPipeError:
			break


if __name__ == '__main__':
	main()



"""
Minimalistic GTF parsing.
"""
from collections import namedtuple
from ..io.xopen import xopen

GtfRecord = namedtuple('GtfRecord',
	['chrom', 'source', 'feature', 'start', 'stop', 'score', 'strand', 'frame',
		'attributes'
	])


# this parser is very simplistic. It is only made to be able to parse ENSEMBL files such as these:
# ftp://ftp.ensembl.org/pub/release-66/gtf/homo_sapiens/Homo_sapiens.GRCh37.66.gtf.gz
#
# First line in that file:
# GL000213.1 protein_coding exon 138767 139339 . - . gene_id "ENSG00000237375"; transcript_id "ENST00000327822";
#    exon_number "1"; gene_name "BX072566.1"; gene_biotype "protein_coding"; transcript_name "BX072566.1-201";


def _none_or_type(s, type):
	return None if s == '.' else type(s)


def GtfReader(path):
	"""Iterate over the GTF file named by given path. Yield GtfRecord objects."""
	with xopen(path) as f:
		for line in f:
			if line.startswith('#'):
				continue
			line = line.rstrip('\n\r')
			fields = line.split('\t')
			chrom = fields[0]
			source = fields[1]
			feature = fields[2]
			start = int(fields[3]) - 1
			stop = int(fields[4])
			score = _none_or_type(fields[5], float)
			strand = fields[6]
			frame = _none_or_type(fields[7], int)
			assert strand in '+-'
			assert frame in (None, 0, 1, 2)

			atts = fields[8].strip(' ').split(';')
			attributes = dict()
			for att in atts:
				att = att.strip(' ')
				if not att:
					continue
				name, value = att.split(' ', maxsplit=1)
				value = value.strip('"')
				attributes[name] = value
			yield GtfRecord(
				chrom=chrom,
				source=source,
				feature=feature,
				start=start,
				stop=stop,
				score=score,
				strand=strand,
				frame=frame,
				attributes=attributes
			)


def main():
	import sys
	for record in parse_gtf(sys.argv[1]):
		print(record)

if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Modify FASTA and FASTQ files by picking subsets and modifying individual entries.

Possible modifications:
- Pick a subset of records (given by name). With lots of names, this is faster
  than 'grep -A 3 --no-group-separator -f readnames.txt file.fastq' magic, which
  may be used with FASTQ files.
  If the record name ends in '/1' or '/2', these two charecter are ignored when
  comparing to the names in the file.
- Trim low-quality ends
- Trim reads to a given length
- Discard reads shorter than a given length
- Discard reads in which the expected number of errors exceeds a threshold
- Discard reads that contain characters other than those in a given set.
- Reverse-complement each read
- Convert from FASTA to FASTQ by assigning a fixed quality value to all bases
- Convert from FASTQ to FASTA by dropping all quality values
- Make read names unique
- Pick only the first N sequences.

Modifications are done in the order in which they are listed above.
The result is written to standard output.

The algorithm for quality trimming is the same as the one used by BWA:
- Subtract the cutoff value from all qualities.
- Compute partial sums from all indices to the end of the sequence.
- Trim sequence at the index at which the sum is minimal.
"""
import sys
import errno
from collections import defaultdict
from itertools import islice
from sqt import HelpfulArgumentParser, SequenceReader, FastaWriter, FastqWriter
from sqt.dna import reverse_complement
from sqt.qualtrim import quality_trim_index as trim_index, expected_errors

__author__ = "Marcel Martin"


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	arg = parser.add_argument
	arg('--names', metavar='FILE', default=None,
		help='Keep only those records whose names occur in FILE (one per line)')
	arg("-q", "--cutoff", type=int, default=None,
		help="Quality cutoff. Only when input format is FASTQ")
	arg("--length", "-l", type=int, default=None,
		help="Shorten records to LENGTH (default: do not shorten)")
	arg('-m', '--minimum-length', type=int, default=0, metavar='LENGTH',
		help='Discard reads shorter than LENGTH')
	arg('--max-errors', type=float, default=None, metavar='ERRORS',
		help='Discard reads whose expected number of errors (computed '
			'from quality values) exceeds ERRORS.')
	arg('--allowed-characters', default=None, metavar='CHARS', help='Discard '
		'reads that contain characters other than those in CHARS. CHARS is '
		'case-sensitive. Example: -c ACGTacgt.')
	arg('--reverse-complement', '-r', action='store_true',
		default=False, help='Reverse-complement each sequence')
	arg('--constant-quality', '-c', metavar='QUALITY', type=int, default=None,
		help='Set all quality values to QUALITY. Use this to convert from '
			'FASTA to FASTQ.')
	arg('--fasta', default=False, action='store_true',
		help='Always output FASTA (drop qualities if input is FASTQ)')
	arg('--unique-names', action='store_true', default=False,
		help="Make record names unique by appending _1, _2 etc. when necessary")
	arg('--limit', '-n', type=int, metavar='N', default=None,
		help="Pick only the first N sequences (default: all)")

	arg("--width", "-w", type=int, default=80,
		help="Characters per line in output FASTA (default: %(default)s). "
			"Set to 0 to disallow line breaks entirely. This is ignored for FASTQ files.")
	arg('path', metavar='FASTA/FASTQ',
		help='input FASTA or FASTQ file')
	"""
	Some possible extensions to this tool:

	arg("--histogram", action="store_true", default=False,
		help="Print a histogram of the length of removed ends")
	arg("-c", "--colorspace", action="store_true", default=False,
		help="Assume input files are in color space and that the sequences contain an initial primer base")
	arg("-z", "--clip-qualities", action="store_true", default=False,
		help="Replace negative qualities with 0. Assumes qualities are stored as chr(qual+33).")
	arg("-n", "--limit", type=int, default=None,
		help="Limit output to N reads (discarding the rest) (default: no limit).")
	"""
	return parser


class ReadPicker:
	def __init__(self, file_with_names):
		read_names = []
		with open(file_with_names) as f:
			read_names = f.read().split('\n')
		self.read_names = { rn for rn in read_names if rn != '' }

	def __call__(self, read):
		rname = read.name.split(' ', maxsplit=1)[0]
		if rname.endswith('/1'):
			rname = rname[:-2]
		elif rname.endswith('/2'):
			rname = rname[:-2]
		if rname in self.read_names:
			return read
		else:
			return None


class QualityTrimmer:
	def __init__(self, cutoff):
		self.cutoff = cutoff

	def __call__(self, read):
		index = trim_index(read.qualities, self.cutoff)
		return read[:index]


class Shortener:
	def __init__(self, length):
		self.length = length

	def __call__(self, read):
		return read[:self.length]


class MinimumLengthFilter:
	def __init__(self, length):
		self.minimum_length = length

	def __call__(self, read):
		if len(read) < self.minimum_length:
			return None
		else:
			return read


class MaxExpectedErrorFilter:
	"""
	Discard reads whose expected number of errors, according to the quality
	values, exceeds the given threshold.

	The idea comes from usearch's -fastq_maxee parameter
	(http://drive5.com/usearch/).
	"""
	def __init__(self, max_errors):
		self.max_errors = max_errors

	def __call__(self, read):
		if expected_errors(read.qualities) > self.max_errors:
			return None
		else:
			return read


class AllowedCharacterFilter:
	"""
	Discard reads that contain characters other than those in the given set.
	"""
	def __init__(self, allowed_characters):
		self.allowed = set(allowed_characters)

	def __call__(self, read):
		if set(read.sequence) <= self.allowed:
			return read
		else:
			return None


def reverse_complementer(read):
	read.sequence = reverse_complement(read.sequence)
	if read.qualities:
		read.qualities = read.qualities[::-1]
	return read


class QualitySetter:
	def __init__(self, value):
		self.quality_char = chr(33 + value)

	def __call__(self, read):
		read.qualities = self.quality_char * len(read)
		return read


def quality_dropper(read):
	read.qualities = None
	return read


class UniqueNamer:
	def __init__(self):
		# Counter for occurrences of a name
		self.names = defaultdict(int)

	def __call__(self, read):
		if ' ' in read.name:
			name, description = read.name.split(' ', maxsplit=1)
		else:
			name = read.name
			description = None
		self.names[name] += 1
		if self.names[name] == 1:
			# Read not previously seen
			return read
		name = '{}_{}'.format(name, self.names[name] - 1)
		read.name = name
		if description is not None:
			read.name += ' ' + description
		return read


def main():
	parser = get_argument_parser()
	args = parser.parse_args()
	if args.width == 0:
		args.width = None

	modifiers = []
	if args.names:
		modifiers.append(ReadPicker(args.names))
	if args.cutoff is not None:
		modifiers.append(QualityTrimmer(args.cutoff))
	if args.length:
		modifiers.append(Shortener(args.length))
	if args.minimum_length != 0:
		modifiers.append(MinimumLengthFilter(args.minimum_length))
	if args.max_errors is not None:
		modifiers.append(MaxExpectedErrorFilter(args.max_errors))
	if args.allowed_characters is not None:
		modifiers.append(AllowedCharacterFilter(args.allowed_characters))
	if args.reverse_complement:
		modifiers.append(reverse_complementer)
	if args.constant_quality is not None:
		modifiers.append(QualitySetter(args.constant_quality))
	if args.fasta:
		modifiers.append(quality_dropper)
	if args.unique_names:
		modifiers.append(UniqueNamer())
	with SequenceReader(args.path) as fr:
		format = fr.format
		outformat = format
		if args.constant_quality is not None:
			outformat = 'fastq'
		if args.fasta:
			outformat = 'fasta'
		if outformat == 'fastq':
			writer = FastqWriter(sys.stdout)
		else:
			writer = FastaWriter(sys.stdout, line_length=args.width)
		try:
			for record in islice(fr, 0, args.limit):
				for modifier in modifiers:
					record = modifier(record)
					if record is None:
						break
				else:
					# only executed if loop did not terminate via break
					writer.write(record)
		except IOError as e:
			if e.errno != errno.EPIPE:
				raise


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
"""
import sys
from random import randint, random, seed
import matplotlib as mpl
mpl.use('pdf')  # enable matplotpib over an ssh connection without X
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sqt import HelpfulArgumentParser
from sqt.io.fasta import IndexedFasta

__author__ = "Marcel Martin"


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	add = parser.add_argument
	add("--left", type=float, default=None)
	add("--right", type=float, default=None)
	add("--title", help="Plot title (default: Histogram of <INFILENAME>)")
	add("--bins", type=int, default=40, help="number of bins (default: %(default)s")
	add("infile", help="use '-' for standard input")
	add("image", nargs='?', help="name of PDF or SVG file")
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()
	if args.infile == '-':
		path = sys.stdin.buffer
		title = 'Histogram'
	else:
		path = args.infile
		title = 'Histogram of ' + args.infile
	if args.title:
		title = args.title
	data = pd.read_csv(path, dtype=float).values  # delimiter=None is default

	fig, ax = plt.subplots()
	ax.set_title(title)
	plt.hist(data, bins=args.bins, rwidth=0.8)
	ax.set_xlim(left=args.left, right=args.right)
	if args.image is not None:
		plt.savefig(args.image)
	else:
		plt.show()


if __name__ == '__main__':
	main()



"""
CIGAR operations.

There are two ways to represent a CIGAR string:
- as a string, such as "17M1D5M4S"
- as a list of (operator, length) pairs, as used by pysam:
[ (0, 17), (1, 2), (0, 5), (0, 4) ]

The naming convention in this module uses cigar and cigar_string to
distinguish both types.

The mapping of CIGAR operators to numbers is:
MIDNSHPX= => 012345678
"""
import sys
from itertools import repeat, chain

__author__ = 'Marcel Martin'

# constants
M = 0  # match or mismatch
I = 1  # insertion
D = 2  # deletion
N = 3  # skipped reference region
S = 4  # soft clipping
H = 5  # hard clipping
P = 6  # padding
X = 7  # mismatch
EQ = 8 # match

# use this as a sequence to map an encoded operation to the appropriate
# character
OPERATORS = 'MIDNSHPX='
DECODE = OPERATORS

# this dictionary maps operations to their integer encodings
_ENCODE = dict( (c,i) for (i, c) in enumerate(DECODE) )


def _assert_at_end(i):
	"""Assert that the iterator i is at its end"""
	if __debug__:
		try:
			next(i)
			assert False
		except StopIteration:
			pass


def alignment_iter(read, ref, cigar, gap='-'):
	"""
	Yield triples (read_char, reference_char, cigar_char) that
	fully describe the alignment betwen read and ref according to cigar.

	If the cigar operation is a 'M', the cigar_char is set to either
	'=' or 'X' depending on whether read_char matches reference_char
	or not.

	At gaps in the alignment, either read_char or reference_char are
	set to the given gap character.

	read -- an iterable representing the read
	ref -- an iterable representing the reference sequence
	cigar -- a list of (operator, length) pairs
	"""
	i = iter(read)
	j = iter(ref)
	for op in decoded_ops(cigar):
		if op == 'M':
			ci = chr(next(i))
			cj = chr(next(j))
			yield (ci, cj, '=' if ci == cj else 'X')
		elif op == 'I':
			yield (chr(next(i)), gap, 'I')
		elif op == 'D':
			yield (gap, chr(next(j)), 'D')
		else:
			raise ValueError("CIGAR operator {} not supported".format(op))
	_assert_at_end(i)
	_assert_at_end(j)


def print_alignment(read, ref, cigar, file=sys.stdout):
	"""
	Print an alignment between read and ref according to a CIGAR.
	This uses the alignment_iter() function from above.

	cigar -- a list of (operator, length) pairs
	"""
	row1 = ''
	row2 = ''
	align = ''
	for read_char, reference_char, op in alignment_iter(read, ref, cigar):
		row1 += read_char
		align += op
		row2 += reference_char
	print(row1, align, row2, sep='\n', file=file)


def unclipped_region(cigar):
	"""
	Return tuple (cigar, start, stop), where cigar is the given cigar without soft clipping
	and (start, stop) is the interval in which the read is *not* soft-clipped.
	"""
	if cigar[0][0] == S:
		start = cigar[0][1]
		cigar = cigar[1:]
	else:
		start = 0
	if cigar[-1][0] == S:
		stop = -cigar[-1][1]
		cigar = cigar[:-1]
	else:
		stop = None
	return (cigar, start, stop)


def reference_to_query_length(cig, reference_bases):
	"""
	Given a prefix of length reference_bases relative to the reference,
	how long is the prefix of the read?

	If the position is within an insertion, then the number of bases up to

	Hard- and soft-clipped bases are always included in the resulting coordinate.
	"""
	rpos = 0
	qpos = 0
	for op, length in cig:
		if op == S or op == H:
			qpos += length
		elif op == M:
			rpos += length
			qpos += length
			if rpos >= reference_bases:
				return qpos + reference_bases - rpos
		elif op == D:
			rpos += length
			if rpos >= reference_bases:
				return qpos
		elif op == I:
			qpos += length
		else:
			raise ValueError('CIGAR operator', op, 'not supported, yet')
	return None


class Cigar:
	"""
	Representation of an alignment in the form of a CIGAR string.

	TODO
	- Rename .cigar attribute to .ops/.opslist/.operations?
	- What should len(Cigar(...)) return? Length of .cigar attribute?
	- Length of alignment on reference
	- What should __iter__ do? Should it be .elements()?
	"""
	def __init__(self, cigar):
		"""
		cigar -- either a string such as '3M2I2M' or a list of
		         (operator, length) tuples.
		"""
		if isinstance(cigar, str):
			self.cigar = self.parse(cigar)
		# TODO
		# elif isinstance(cigar, Cigar): self.cigar = cigar.cigar
		else:
			self.cigar = cigar

	def __eq__(self, other):
		return self.cigar == other.cigar

	def __ne__(self, other):
		return self.cigar != other.cigar

	def _as_string(self, join_by=''):
		"""
		Format the CIGAR string.

		join_by is an optional separator.

		>>> Cigar('3M2S')._as_string(join_by=' ')
		'3M 2S'
		"""
		return join_by.join(
			'{}{}'.format(l, DECODE[op]) for op, l in self.cigar)

	def __format__(self, format_spec):
		if format_spec in ('', ' '):
			return self._as_string(join_by=format_spec)
		else:
			raise ValueError(
				"Format specification '{}' not supported".format(format_spec))

	def __str__(self):
		return self._as_string()

	def __repr__(self):
		return "Cigar('{}')".format(str(self))

	@staticmethod
	def parse(cigar_string):
		"""
		Parse CIGAR string and return a list of (operator, length) pairs. Spaces
		are ignored.

		>>> parse("3S17M8D4M9I3H")
		[(4, 3), (0, 17), (2, 8), (0, 4), (1, 9), (5, 3)]

		>>> parse("3S 17M")
		[(4, 3), (0, 17)]
		"""
		cigar = []
		n = ''  # This is a string to which digits are appended
		for c in cigar_string:
			if c.isdigit():
				n += c
			elif c in _ENCODE:
				if n == '':
					raise ValueError("CIGAR string should start with a number.")
				cigar.append( (_ENCODE[c], int(n)) )
				n = ''
			elif c == ' ':
				continue
			else:
				raise ValueError(
					'Character "{}" unexpected in CIGAR string.'.format(c))
		if n != '':
			raise ValueError("Unexpected end of CIGAR string.")
		return cigar

	def __add__(self, other):
		"""
		Return the concatenation of this CIGAR and another one.

		>>> Cigar('2S1M') + Cigar('3M4S'))
		Cigar('2S4M4S')
		"""
		if len(other.cigar) == 0:
			return Cigar(self.cigar)
		if len(self.cigar) == 0:
			return Cigar(other.cigar)
		self_last = self.cigar[-1]
		other_first = other.cigar[0]
		# same operation?
		if self_last[0] == other_first[0]:
			return Cigar(
				self.cigar[:-1] +
				[(self_last[0], self_last[1] + other_first[1])] +
				other.cigar[1:])
		return Cigar(self.cigar + other.cigar)

	def elements(self, numbers=False):
		"""
		Yield all operations one by one.

		If numbers is set to True, the operations are returned numerically.

		>>> ''.join(Cigar("3S2I3M").elements())
		"SSSIIMMM"

		>>> list(Cigar("3S2I3M").elements(numbers=True))
		[4, 4, 4, 1, 1, 0, 0, 0]
		"""
		if numbers:
			return chain.from_iterable(repeat(op, l) for (op, l) in self.cigar)
		else:
			return chain.from_iterable(
				repeat(DECODE[op], l) for (op, l) in self.cigar)

	def alignment_length(self):
		"""
		Return the number of bases of the read that are used in the alignment.
		This counts all matches, mismatches and insertions. Clipped bases are
		not counted.
		"""
		return sum(l for op, l in self.cigar if op in (M, I, EQ, X))

	def query_length(self, count_clipped='soft'):
		"""
		Return the length of the query sequence deduced from the
		length of the operations used in this CIGAR string.
		Matches, mismatches and insertions are counted (M, I, =, X operators).

		The count_clipped parameter determines whether hard- or soft-clipped
		bases are counted (H and S operators). It can be set to None, 'soft' or
		'hard'.

		- None: Do not count any clipped bases. Only the bases actually
			aligned to the reference are counted.

		- 'soft': Count also soft-clipped bases. The returned length should be
			identical to the length of the SEQ field in a SAM file that
			follows the specification. From the spec: 'sum of lengths of the
			M/I/S/=/X operations shall equal the length of SEQ'.

		- 'hard': Count both soft- and hard-clipped bases. The returned length
			is the same as the length of the original read.
		"""
		if count_clipped is None:
			ops_to_count = (M, I, EQ, X)
		elif count_clipped == 'soft':
			ops_to_count = (M, I, EQ, X, S)
		elif count_clipped == 'hard':
			ops_to_count = (M, I, EQ, X, S, H)
		else:
			raise ValueError(
				"count_clipped must be either None, 'soft' or 'hard'")

		return sum(length for op, length in self.cigar if op in ops_to_count)

	def reference_length(self):
		"""
		Return the length of the reference sequence deduced from the length of
		the operations used in the CIGAR string. This counts the M, D, N, X, =
		operations.
		"""
		return sum(length for op, length in self.cigar if op in
			 (M, D, EQ, X, N))

# deprecated functions that have become methods of the Cigar class

def decoded_ops(cigar):
	return Cigar(cigar).elements()


def ops(cigar):
	return Cigar(cigar).elements(numbers=True)


def as_string(cigar, join_by=''):
	return str(Cigar(cigar))


def parse(cigar_string):
	return Cigar.parse(cigar_string)


def concat(left, right):
	return Cigar(left) + Cigar(right)


def aligned_bases(cigar):
	return Cigar(cigar).query_length(count_clipped=None)


def seq_length(cigar):
	return Cigar(cigar).query_length(count_clipped='soft')


def read_length(cigar):
	return Cigar(cigar).query_length(count_clipped='hard')



#!/usr/bin/env python3
"""
- Python 2 and 3 compatible fast reverse complement
- definition of genetic code (triplets to amino acids)
- other misc. functions
"""
import sys

if sys.version < '3':
	from string import maketrans
else:
	maketrans = bytes.maketrans
	_TR_STR = str.maketrans('ACGTUMRWSYKVHDBNacgtumrwsykvhdbn', 'TGCAAKYWSRMBDHVNtgcaakywsrmbdhvn')

_TR = maketrans(b'ACGTUMRWSYKVHDBNacgtumrwsykvhdbn', b'TGCAAKYWSRMBDHVNtgcaakywsrmbdhvn')

if sys.version < '3':
	def reverse_complement(s):
		return s.translate(_TR)[::-1]
else:
	def reverse_complement(s):
		if isinstance(s, str):
			return s.translate(_TR_STR)[::-1]
		else:
			return s.translate(_TR)[::-1]


GENETIC_CODE = {
	'AAA': 'K',
	'AAC': 'N',
	'AAG': 'K',
	'AAT': 'N',
	'ACA': 'T',
	'ACC': 'T',
	'ACG': 'T',
	'ACT': 'T',
	'AGA': 'R',
	'AGC': 'S',
	'AGG': 'R',
	'AGT': 'S',
	'ATA': 'I',
	'ATC': 'I',
	'ATG': 'M',
	'ATT': 'I',
	'CAA': 'Q',
	'CAC': 'H',
	'CAG': 'Q',
	'CAT': 'H',
	'CCA': 'P',
	'CCC': 'P',
	'CCG': 'P',
	'CCT': 'P',
	'CGA': 'R',
	'CGC': 'R',
	'CGG': 'R',
	'CGT': 'R',
	'CTA': 'L',
	'CTC': 'L',
	'CTG': 'L',
	'CTT': 'L',
	'GAA': 'E',
	'GAC': 'D',
	'GAG': 'E',
	'GAT': 'D',
	'GCA': 'A',
	'GCC': 'A',
	'GCG': 'A',
	'GCT': 'A',
	'GGA': 'G',
	'GGC': 'G',
	'GGG': 'G',
	'GGT': 'G',
	'GTA': 'V',
	'GTC': 'V',
	'GTG': 'V',
	'GTT': 'V',
#	'TAA': stop
	'TAC': 'Y',
#	'TAG': stop,
	'TAT': 'Y',
	'TCA': 'S',
	'TCC': 'S',
	'TCG': 'S',
	'TCT': 'S',
#	'TGA': stop
	'TGC': 'C',
	'TGG': 'W',
	'TGT': 'C',
	'TTA': 'L',
	'TTC': 'F',
	'TTG': 'L',
	'TTT': 'F'
}


def n_intervals(sequence, N='N'):
	"""
	Given a sequence, yield all intervals containing only N characters as
	tuples (start, stop). If the sequence is a bytes/bytearray object,
	set N=ord(b'N')

	>>> list(n_intervals('ACGTnNAC'))
	[(4, 6)]
	>>> list(n_intervals(b'ACGTnNAC', N=ord(b'N')))
	[(4, 6)]
	"""
	sequence = sequence.upper()
	start = sequence.find(N)
	while start >= 0:
		stop = start + 1
		while stop < len(sequence) and sequence[stop] == N:
			stop += 1
		yield (start, stop)
		start = sequence.find(N, stop)


def intervals_complement(intervals, length):
	"""
	Given an iterable of sorted, nonoverlapping intervals as (start, stop)
	pairs, yield the complementary intervals. The result is equivalent to
	[(0, length)] minus the given intervals.

	>>> list(intervals_complement([(1, 2), (4, 6)], length=10))
	[(0, 1), (2, 4), (6, 10)]
	"""
	prev_stop = 0
	for start, stop in intervals:
		if start >= length:
			break
		if prev_stop != start:
			yield (prev_stop, start)
		prev_stop = stop
	if prev_stop < length:
		yield (prev_stop, length)



#!/usr/bin/env python3
"""

"""
__author__ = 'Marcel Martin'

import sys
from collections import Counter
from sqt import HelpfulArgumentParser
from pysam import Samfile


def main():
	parser = HelpfulArgumentParser(description=__doc__)
#	parser.add_argument('--sorted', '-s', action='store_true', default=False)
	parser.add_argument('--quality', '-q', type=int, default=-1000,
		help='minimum mapping quality')
	parser.add_argument('--minimum-start', '-m', type=int, default=1,
		help='minimum start position (1-based). Reads starting earlier are ignored.')
	parser.add_argument('--maximum-start', '-M', type=int, default=1E100,
		help='maximum start position (1-based). Reads starting later are ignored.')
	parser.add_argument('--forward', action='store_true', default=False,
		help='count only reads mapped forward')
	parser.add_argument("bam", metavar="SAM/BAM", help="name of a SAM or BAM file")
	args = parser.parse_args()

	allexpressions = []
	counter = Counter()
	coverage = Counter()
	with Samfile(args.bam) as samfile:
		for read in samfile:
			if read.mapq < args.quality or read.is_unmapped:
				continue
			if args.forward and read.is_reverse:
				continue
			if not (args.minimum_start <= read.pos + 1 <= args.maximum_start):
				continue
			counter[read.rname] += 1

			# TODO should qlen or alen be used?
			# qlen: no. of aligned bases in the read
			# alen: no. of aligned bases in the reference
			coverage[read.rname] += read.qlen
		reference_names = samfile.references

	for index, name in enumerate(reference_names):
		print(name, counter[index], coverage[index], sep='\t')

if __name__ == '__main__':
	main()



def frequency_median(frequencies):
	"""
	Given a dictionary of frequencies, return the median.
	If the total no. of values is odd, the left of both
	middle values is returned.
	"""
	m = 0  # partial sum
	middle = (1 + sum(frequencies.values())) // 2
	for length in sorted(frequencies):
		m += frequencies[length]
		if m >= middle:
			return length
	# never reached
	assert False



#!/usr/bin/env python3
"""
Compute per-reference statistics given a BAM file and a matching FASTA reference.

The output is a tab-separated table with the following columns:

chromosome:
  reference/scaffold/chromosome name

gc:
  GC content (number of found G, C bases divided by the number of
  found A, C, G, T bases)

length:
  length of the reference

non_acgt:
  number of characters that are not one of A, C, G, T

bases:
  total number of bases aligned to the reference, not counting those aligning
  to bases that are neither A, C, G nor T

avg_cov:
  average coverage = bases / (length - non_acgt)

median_cov:
  median coverage (as above, coverage over non-ACGT bases is not counted)

TODO:
- optionally print summary over all reference sequences
"""
__author__ = 'Marcel Martin'

import sys
from collections import Counter, namedtuple
from pysam import Samfile
from sqt import HelpfulArgumentParser
from sqt._helpers import byte_frequencies
from sqt.math import frequency_median
from sqt.io.fasta import IndexedFasta

Info = namedtuple('Info', 'length acgt_length bases acgt_bases median_coverage acgt_median_coverage gc')


def collect_info(samfile, fasta, name, mask):
	sequence = fasta.get(name)[:]
	length = len(sequence)
	freqs = byte_frequencies(sequence)
	assert length == samfile.lengths[samfile.gettid(name)]
	acgt_length = sum(freqs[c] for c in b'ACGTacgt')
	gc = sum(freqs[c] for c in b'GCgc')

	# Get coverage
	acgt_bases = 0
	bases = 0
	coverages = Counter()
	acgt_coverages = Counter()
	for column in samfile.pileup(name, stepper='all', mask=mask):
		n = column.n
		if sequence[column.pos] in b'ACGTacgt':
			acgt_bases += n
			acgt_coverages[n] += 1
		bases += n
		coverages[n] += 1

	# Compute medians
	assert coverages[0] == 0
	assert acgt_coverages[0] == 0
	coverages[0] = length - sum(coverages.values())
	acgt_coverages[0] = acgt_length - sum(acgt_coverages.values())

	median_coverage = frequency_median(coverages)
	acgt_median_coverage = frequency_median(acgt_coverages)

	return Info(
		length=length,
		acgt_length=acgt_length,
		bases=bases,
		acgt_bases=acgt_bases,
		median_coverage=median_coverage,
		acgt_median_coverage=acgt_median_coverage,
		gc=gc,
	)


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument('--mask', default='0x504',
		help="Exclude reads in which flags have one of the bits in MASK set. "
			"Default is 0x504, that is, exclude unmapped reads (0x4), "
			"secondary alignments (0x100) and PCR/optical duplicates (0x400).")
	parser.add_argument("fasta", metavar="FASTA", help="path to reference FASTA file")
	parser.add_argument("bam", metavar="BAM", help="path to BAM file")
	args = parser.parse_args()
	try:
		args.mask = int(args.mask, base=0)
	except ValueError:
		parser.error('Could not interpret mask value "{}"'.format(args.mask))

	print('chromosome', 'gc', 'length', 'non_acgt', 'bases', 'avg_cov', 'median_cov', sep='\t')
	with Samfile(args.bam) as samfile, IndexedFasta(args.fasta) as fasta:
		reference_names = samfile.references
		for name in reference_names:
			info = collect_info(samfile, fasta, name, args.mask)
			print(
				name,
				"{:.4f}".format(info.gc / info.acgt_length),
				info.length,
				info.length - info.acgt_length,
				info.acgt_bases,
				"{:.3f}".format(info.acgt_bases / info.acgt_length),
				info.acgt_median_coverage,
				sep='\t',
			)


if __name__ == '__main__':
	main()



"""
ANSI escape codes and some helper functions for colored terminal output.
"""

# dark foreground colors
BLACK = "\x1b[0;30m"
RED = "\x1b[0;31m"
GREEN = "\x1b[0;32m"
BROWN = "\x1b[0;33m"
BLUE = "\x1b[0;34m"
PURPLE = "\x1b[0;35m"
CYAN = "\x1b[0;36m"
GRAY = "\x1b[0;37m"

# light foreground colors
DARKGRAY = "\x1b[1;30m"
LIGHTRED = "\x1b[1;31m"
LIGHTGREEN = "\x1b[1;32m"
YELLOW = "\x1b[1;33m"
LIGHTBLUE = "\x1b[1;34m"
LIGHTPURPLE = "\x1b[1;35m"
LIGHTCYAN = "\x1b[1;36m"
WHITE = "\x1b[1;37m"

# dark background colors
BACKGROUND_BLACK = "\x1b[0;40m"
BACKGROUND_BLUE = "\x1b[0;44m"
BACKGROUND_GREEN = "\x1b[0;42m"
BACKGROUND_CYAN = "\x1b[0;46m"
BACKGROUND_RED = "\x1b[0;41m"
BACKGROUND_PURPLE = "\x1b[0;45m"
BACKGROUND_BROWN = "\x1b[0;43m"
BACKGROUND_GRAY = "\x1b[0;47m"

# light background colors
BACKGROUND_DARKGRAY = "\x1b[1;40m"
BACKGROUND_LIGHTBLUE = "\x1b[1;44m"
BACKGROUND_LIGHTGREEN = "\x1b[1;42m"
BACKGROUND_LIGHTCYAN = "\x1b[1;46m"
BACKGROUND_LIGHTRED = "\x1b[1;41m"
BACKGROUND_LIGHTPURPLE = "\x1b[1;45m"
BACKGROUND_YELLOW = "\x1b[1;43m"
BACKGROUND_WHITE = "\x1b[1;47m"

# other codes
RESET = "\x1b[0m"

BOLD_ON = "\x1b[1m"
ITALICS_ON = "\x1b[3m"
UNDERLINE_ON = "\x1b[4m"
INVERSE_ON = "\x1b[7m"
STRIKETHROUGH_ON = "\x1b[9m"

BOLD_OFF = "\x1b[22m"
ITALICS_OFF = "\x1b[23m"
UNDERLINE_OFF = "\x1b[24m"
INVERSE_OFF = "\x1b[27m"
STRIKETHROUGH_OFF = "\x1b[29m"


def colored(s, color):
	"""
	Enclose string s in ANSI escape codes such that s appears in the given
	color when printed on a terminal.
	"""
	return color + s + RESET


def red(s):
	return colored(s, RED)


def blue(s):
	return colored(s, BLUE)


def green(s):
	return colored(s, GREEN)


def yellow(s):
	return colored(s, YELLOW)


def lightred(s):
    return colored(s, LIGHTRED)


def bgred(s):
	return colored(s, BACKGROUND_RED)


def bgblue(s):
	return colored(s, BACKGROUND_BLUE)


def bggreen(s):
	return colored(s, BACKGROUND_GREEN)

bgreen = bggreen

def bgyellow(s):
	return colored(s, BACKGROUND_YELLOW)



#!/usr/bin/env python3
# kate: word-wrap-column 80; word-wrap off;
"""
Print a report about a SAM/BAM file.
"""
__author__ = 'Marcel Martin'

import sys
from collections import Counter, namedtuple, defaultdict
from itertools import islice
from contextlib import ExitStack
from pysam import Samfile
from sqt import HelpfulArgumentParser, Cigar, cigar


Subalignment = namedtuple("Subalignment", ['length', 'pos', 'tid'])

class SupplementaryAlignment(namedtuple("SupplementaryAlignment",
	['rname', 'pos', 'strand', 'cigar', 'mapq', 'edit_distance'])):
	# TODO cigar attribute should be a Cigar object
	@property
	def reference_name(self):
		return self.rname

	@property
	def qstart(self):  # TODO rename to query_start
		return soft_clipping_length(self.cigar, 0)

	@property
	def qend(self):   # TODO rename to query_end
		return soft_clipping_length(self.cigar, 0) + cigar.aligned_bases(self.cigar)

	@property
	def qlength(self):
		return cigar.aligned_bases(self.cigar)

	@property
	def rstop(self):
		return self.pos + Cigar(self.cigar).reference_length()


def header(s):
	print(s)
	print('-' * len(s))
	print()


def soft_clipping_length(cig, where):
	"""
	Return length of soft-clipping at the ends of an alignment.

	where -- 0 stands for the beginning of alignment, -1 for the end.
	"""
	assert where in (0, -1)
	if not cig:
		return 0
	p = cig[where]
	if p[0] == cigar.S:
		return p[1]
	return 0



class AlignedRead:
	"""
	An AlignedRead describes all alignments that exist for a single read in a
	SAM/BAM file. Thus, it also includes all supplementary alignments (if they
	exist).
	"""
	def __init__(self, read, reference_name):
		"""
		The given read should be an AlignedSegment. The SA tag of the read is
		parsed.
		"""
		assert not read.is_secondary, "Read should not be secondary"
		assert not read.is_supplementary, "Read should not be supplementary"
		self.alignments = extract_supplementary_alignments(read, reference_name)
		self.query_name = read.query_name
		self._primary_read = read
		self._length = Cigar(read.cigar).query_length(count_clipped='hard')
		assert self._length == Cigar(read.cigar).query_length(count_clipped='soft'), "Shouldn't be hard-clipped"

	def __len__(self):
		return self._length

	def overall_aligned_bases(self):
		return overall_aligned_bases(self._primary_read, self.alignments)


def parse_supplementary(sa):
	"""
	Parse supplementary alignments given by the SA:Z: tag in a SAM record.

	Return a list of SupplementaryAlignment objects.
	"""
	fields = sa.split(';')
	assert fields[-1] == ''  # sa must end with a ';'
	alignments = []
	for field in fields[:-1]:
		ref, pos, strand, cig, mapq, edit_dist = field.split(',')
		pos = int(pos) - 1
		cig = Cigar.parse(cig)  # TODO should be Cigar(cig)
		mapq = int(mapq)
		edit_dist = int(edit_dist)
		assert strand in '+-'
		a = SupplementaryAlignment(ref, pos, strand, cig, mapq, edit_dist)
		alignments.append(a)
	return alignments


def extract_supplementary_alignments(read, rname):
	"""
	Given a single read that potentially has supplementary alignments specified
	by the SA tag, return a list of SupplementaryAlignment objects. The list
	includes at least the read itself.
	"""
	tags = dict(read.tags)
	alignments = [SupplementaryAlignment(rname, read.pos, '+-'[int(read.is_reverse)],
		read.cigar, read.mapq, tags['NM'])]
	if 'SA' in tags:
		alignments.extend(parse_supplementary(tags['SA']))
	return alignments


def print_coverage_report(aligned_read, report, minimum_cover_fraction=0.01):
	"""
	Print coverage report for a single read to the file-like object 'report'.
	"""
	read_length = len(aligned_read)
	alignments = aligned_read.alignments
	print('Read {} ({:.1f}kbp):'.format(aligned_read.query_name, read_length / 1000), file=report)
	for alignment in sorted(alignments, key=lambda a: (a.qstart, a.qend)):
		cigar = Cigar(alignment.cigar)
		alignment_length = cigar.query_length(count_clipped=None)
		if alignment_length / read_length < minimum_cover_fraction:
			continue
		print(
			'{:9} bp'.format(alignment_length),
			'{:6.1%}'.format(alignment_length / read_length),
			'{:6} ... {:6}  '.format(alignment.qstart+1, alignment.qend),
			'{} {:>2}:{}-{}'.format(alignment.strand,
				alignment.rname, alignment.pos+1, alignment.rstop),
			file=report)
	bases = aligned_read.overall_aligned_bases()
	print('{:.1%} aligned ({}/{})\n'.format(bases / read_length,
		bases, read_length),
		file=report)


def overall_aligned_bases(read, alignments):
	"""
	Given an AlignedRead and its corresponding list of SupplementaryAlignment
	objects, return how many of its bases are aligned,
	considering all supplementary alignments.
	"""
	events = []
	for alignment in alignments:
		events.append((alignment.qstart, 'start', None))
		events.append((alignment.qend, 'stop', alignment))

	depth = 0  # number of observed 'start' events
	bases = 0  # number of covered bases
	last_qstart = None
	for qpos, what, alignment in sorted(events, key=lambda x: x[0]):
		if what == 'start':
			if depth == 0:
				last_qstart = qpos
			depth += 1
		elif what == 'stop':
			depth -= 1
			if depth == 0:
				# interval (last_qstart, qpos) was covered
				bases += qpos - last_qstart
	return bases


def print_basics(aligned_reads, aligned_bases):
	header('Basic statistics')
	bases = sum(len(aligned_read) for aligned_read in aligned_reads)
	total_reads = len(aligned_reads)
	total_alignments = sum(len(aligned_read.alignments) for aligned_read in aligned_reads)
	print('Number of reads:     {:10,d}'.format(total_reads))
	#print('unfiltered+CIGAR:    {:10,d} ({:.2%})'.format(mapped_reads, mapped_reads / total_reads))
	print('Alignments per read: {:.3f} (only mapped reads)'.format(total_alignments / total_reads))  # TODO is this correct?
	print('bases:         {:15,d} ({:.2f} Gbp)'.format(bases, bases / 1E9))
	print('aligned bases: {:15,d} ({:.2f} Gbp) ({:.2%})'.format(aligned_bases, aligned_bases/1E9, aligned_bases/bases))
	print()


def print_subalignment_histogram(number_alignments):
	print('Histogram of number of subalignments')
	rest = 0
	for number, count in number_alignments.items():
		if number <= 10:
			print(' {:2} {:9}'.format(number, count))
		else:
			rest += count
	if rest > 0:
		print('>10 {:9}'.format(rest))
	print()


def print_subalignment_stats(aligned_reads, total_reads):
	header('Subalignment statistics')
	fully_aligned_95 = 0  # reads whose bases are 95% aligned within one subalignment
	fully_aligned_99 = 0  # reads whose bases are 99% aligned within one subalignment
	number_alignments = Counter()
	interesting = 0
	for aligned_read in aligned_reads:
		alignments = aligned_read.alignments
		lengths = sorted(alignment.qlength for alignment in alignments)
		refnames = set(alignment.reference_name for alignment in alignments)

		if len(lengths) >= 1 and lengths[-1] >= 0.95 * len(aligned_read):
			fully_aligned_95 += 1
		if len(lengths) >= 1 and lengths[-1] >= 0.99 * len(aligned_read):
			fully_aligned_99 += 1
		number_alignments[len(lengths)] += 1

		# is this an 'interesting' read? (arbitrary thresholds)
		if 2 <= len(lengths) <= 4 and len(set(refnames)) > 1:
			interesting += 1

	print_subalignment_histogram(number_alignments)
	print('fully aligned (95%):{:10,d} ({:.2%})'.format(fully_aligned_95, fully_aligned_95/total_reads))
	print('fully aligned (99%):{:10,d} ({:.2%})'.format(fully_aligned_99, fully_aligned_99/total_reads))
	print('no of interesting reads:', interesting)
	print()


def print_cigar_usage(counter):
	header("CIGAR operator usage")
	total_ops = sum(counter.values())
	ops = 'MIDNSHPX='
	for op_i, op in enumerate(ops):
		print("{:2} {:14,d} ({:7.2%})".format(op, counter[op_i], counter[op_i]/total_ops))
	print()


def print_reference_usage(reflengths, reference_hits, minimum_reference_length=1000):
	header('Scaffold/chromosome/references usage')

	long_refs = sum(1 for length in reflengths.values() if length >= minimum_reference_length)
	ref_hits_length = sum(reflengths[refname] for refname in reference_hits)
	total_ref_length = sum(reflengths.values())

	print('total length of references: {:,d} ({:.2f} Gbp)'.format(total_ref_length, total_ref_length/1E9))
	print('references:', len(reflengths))
	print('references hit by at least one alignment:', len(reference_hits))
	print('length of those references: {:,d} ({:.2%})'.format(ref_hits_length, ref_hits_length/total_ref_length))
	print('length of references not hit: {:,d}'.format(total_ref_length - ref_hits_length))
	#_refs = infile.references
	#assert infile.nreferences == len(_refs)
	#refname_to_length = dict(zip(_refs, infile.lengths))
	#for i in range(infile.nreferences):
		#assert refname_to_length[infile.getrname(i)] == reflengths[i]

	long_ref_hits = sum(1 for tid in reference_hits if reflengths[tid] >= minimum_reference_length)
	ref_hits_length = sum(reflengths[refname] for refname in reference_hits)
	print('references >= {} bp:'.format(minimum_reference_length), long_refs)
	print('references >= {} bp hit by at least one alignment:'.format(minimum_reference_length), long_ref_hits)


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument('--quality', '-q', type=int, default=0,
		help='Minimum mapping quality (default: %(default)s')
	parser.add_argument('--minimum-reference-length', metavar='N', type=int, default=0,
		help='For reference usage statistics, ignore references shorter than N.')
	parser.add_argument('--limit', metavar='N', type=int, default=None,
		help='Process only the first N entries in the input file.')
	parser.add_argument('--cover', metavar='FILE', default=None,
		help='Print report about "read coverage" (which sections are aligned) to FILE')
	parser.add_argument('--minimum-cover-fraction', metavar='FRACTION', type=float, default=0.01,
		help='Alignment must cover at least FRACTION of the read to appear in the cover report. (%(default)s)')
	parser.add_argument("bam", metavar="SAM/BAM", help="Name of a SAM or BAM file")
	args = parser.parse_args()

	# Count how often each CIGAR operator occurs
	cigar_counter = Counter()

	n_records = 0
	unmapped = 0
	unmapped_bases = 0
	aligned_reads = []
	with Samfile(args.bam) as sf:
		for record in islice(sf, 0, args.limit):
			n_records += 1
			if record.is_unmapped:
				unmapped += 1
				unmapped_bases += len(record.seq)
				continue
			if record.mapq < args.quality:
				continue
			assert record.cigar is not None
			for op_i, l in record.cigar:
				cigar_counter[op_i] += l

			if not record.is_secondary and not record.is_supplementary:
				aligned_read = AlignedRead(record, sf.getrname(record.tid))
				aligned_reads.append(aligned_read)

		reflengths = sf.lengths
		nreferences = sf.nreferences
		assert nreferences == len(reflengths)
		refnames_map = { tid: sf.getrname(tid) for tid in range(nreferences) }
		reference_lengths = dict(zip(sf.references, sf.lengths))

	total_aligned_bases = 0
	reference_hits = defaultdict(int)
	for aligned_read in aligned_reads:
		total_aligned_bases += aligned_read.overall_aligned_bases()
		for alignment in aligned_read.alignments:
			reference_hits[alignment.reference_name] += 1

	if args.cover is not None:
		with open(args.cover, 'wt') as cover:
			for aligned_read in aligned_reads:
				print_coverage_report(aligned_read, cover, args.minimum_cover_fraction)

	header('All entries in input file')
	print('Total entries:   {:10,d}'.format(n_records))
	print('Unmapped:        {:10,d}'.format(unmapped))
	print('Unmapped bases:  {:10,d}'.format(unmapped_bases))
	print()

	print_basics(aligned_reads, total_aligned_bases)
	print_subalignment_stats(aligned_reads, len(aligned_reads))  # TODO is len(aligned_reads) actually the number of unique reads?
	print_cigar_usage(cigar_counter)
	print_reference_usage(reference_lengths, reference_hits, minimum_reference_length=args.minimum_reference_length)


if __name__ == '__main__':
	main()



"""
Old code that used to be in bamstats:

parser.add_argument('--minimum-start', '-m', type=int, default=1,
	help='minimum start position (1-based). Reads starting earlier are ignored.')
parser.add_argument('--maximum-start', '-M', type=int, default=1E100,
	help='maximum start position (1-based). Reads starting later are ignored.')
parser.add_argument('--forward', action='store_true', default=False,
	help='count only reads mapped forward')


for read in samfile:
	if not (args.minimum_start <= read.pos + 1 <= args.maximum_start):
		continue
	counter[read.rname] += 1

	# TODO should qlen or alen be used?
	# qlen: no. of aligned bases in the read
	# alen: no. of aligned bases in the reference
	coverage[read.rname] += read.qlen
reference_names = samfile.references

for index, name in enumerate(reference_names):
	print(name, counter[index], coverage[index], sep='\t')
"""



#!/usr/bin/env python3
"""
Mutate nucleotides in a FASTA file. Write modified sequences to standard output.
"""
import sys
import random
from sqt import HelpfulArgumentParser
from sqt.io.fasta import FastaReader, FastaWriter

__author__ = "Marcel Martin"


def mutate_sequence(seq, rate=0.1, alphabet='ACGT', indels=True, indel_rate=0.1):
	"""
	TODO this model is not very sophisticated
	"""
	other_chars = {}
	for c in alphabet:
		other_chars[c] = alphabet.replace(c, '')
	mutated = []
	n_sub = 0
	n_indel = 0
	for c in seq:
		c = c.upper()
		r = random.random()
		if r < rate:
			# mutate base
			d = random.choice(other_chars.get(c, 'ACGT'))
			mutated.append(d)
			n_sub += 1
		elif indels and r < (rate + 0.5 * indel_rate):
			# insertion
			mutated.append(random.choice(alphabet))
			mutated.append(c)
			n_indel += 1
		elif indels and r < (rate + indel_rate):
			n_indel += 1
			# deletion
			pass
		else:
			# no change
			mutated.append(c)
	return ''.join(mutated), n_sub, n_indel


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("--rate", type=float, default=0.03,
		help="Substitution rate (default: %(default)s)")
	parser.add_argument("--indel-rate", type=float, default=0.0005,
		help="Indel rate (default: %(default)s)")
	parser.add_argument("--seed", type=int, default=None,
		help="Set random seed for reproducible runs (default: use different seed each run)")
	parser.add_argument("fasta", metavar='FASTA-file', help="Input FASTA file")
	args = parser.parse_args()

	if args.seed is not None:
		random.seed(args.seed)
	fasta_output = FastaWriter(sys.stdout, line_length=80)
	for record in FastaReader(args.fasta):
		mutated, n_sub, n_indel = mutate_sequence(record.sequence, rate=args.rate, alphabet='ACGT', indels=True, indel_rate=args.indel_rate)
		fasta_output.write(record.name + '-sub{}-indel{}'.format(n_sub, n_indel), mutated)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Check whether the EOF marker is present in BAM files.
If it's not, this may be a sign that the BAM file was corrupted.

The exit code is 1 if the marker was present in *all files*.
It is 0 if the marker was missing in any of the files.

BUGS/TODO
- Does not work with uncompressed BAM files.
"""

import sys
from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def bam_eof_is_ok(f):
	"""
	Check whether BAM file f contains the 'magic' end-of-file marker.
	Adapted from samtools function bgzf_check_EOF (in bgzf.c).
	"""
	try:
		f.seek(-28, 2)
	except IOError:
		return False
	data = f.read(28)
	return data == b"\037\213\010\4\0\0\0\0\0\377\6\0\102\103\2\0\033\0\3\0\0\0\0\0\0\0\0\0"


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("-q", "--quiet", action='store_true',
		help="Don't print anything, just set the exit code.")
	parser.add_argument("bam", metavar='BAM', nargs='+')
	args = parser.parse_args()
	exitcode = 0
	for name in args.bam:
		with open(name, 'rb') as f:
			if bam_eof_is_ok(f):
				if not args.quiet:
					print(name, ": OK", sep='')
			else:
				if not args.quiet:
					print(name, ": MISSING", sep='')
					exitcode = 1
				else:
					sys.exit(1)
	sys.exit(exitcode)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Read FASTA-formatted data and replace characters in <FROM-CHARACTERS>
with corresponding characters in <TO-CHARACTERS>.

If the name of a FASTA file is not given, read from standard input.

This is similar to the unix 'tr' command, except that FASTA
comment lines remain unchanged.

You can also translate the sequences within FASTQ files by setting
--format=fastq.

Example (replace C with T and c with t):

%(prog)s translate Cc Tt input.fa > output.fa
"""
import sys
#import string
from sqt import HelpfulArgumentParser
from sqt.io.fasta import FastqReader, FastqWriter, FastaWriter, FastaReader

__author__ = "Marcel Martin"


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("--format", choices=('fasta', 'fastq'), default='fasta')
	parser.add_argument("fromchars", metavar="FROM-CHARACTERS")
	parser.add_argument("tochars", metavar="TO-CHARACTERS")
	parser.add_argument("file", nargs='?', default='-')
	args = parser.parse_args()
	trans = bytes.maketrans(args.fromchars.encode('ascii'), args.tochars.encode('ascii'))

	if args.format == 'fasta':
		fw = FastaWriter(sys.stdout)
		with FastaReader(args.file) as fr:
			for record in fr:
				record.sequence = record.sequence.translate(trans)
				fw.write(record)
	else:
		fw = FastqWriter(sys.stdout)
		with FastqReader(args.file) as fr:
			for record in fr:
				record.sequence = record.sequence.translate(trans)
				fw.write(record)


if __name__ == '__main__':
	main()



import sys

__author__ = "Johannes Koester"

class Interval:
	""" Container for an interval with start and end point. """
	def __init__(self, start, end):
		self.start = start
		self.end = end

	def is_overlap(self, other):
		""" Return True if this interval overlaps with the given interval. """
		return (self.start - other.end <= 0 and self.end - other.start >= 0) or (self.end - other.start >= 0 and self.start - other.end <= 0)
	
	def __repr__(self):
		return "({},{})".format(self.start, self.end)

class IntervalNode:
	"""
	Node in the IntervalTree.
	"""
	def __init__(self, interval, obj, tree):
		self.interval = interval
		self.obj = obj
		self.right = None
		self.left = None
		self.maxend = self.interval.end
		self.height = 0
		self.tree = tree

	def recalc_maxend(self):
		self.maxend = self.interval.end
		if child in (self.left, self.right):
			if child:
				self.maxend = max(child.maxend, self.maxend)

	def __repr__(self):
		return str(self.interval)

class IntervalTree:
	"""
	Interval tree as proposed by Cormen et al.
	"""
	def __init__(self):
		self.root = None

	def insert(self, *args, obj = None):
		"""
		Insert a new interval into the tree. *args can be a start and endpoint given as ints, or a single object with the attributes start and end.
		"""
		node = self._convert_args(*args, obj = obj, node = True)
		self._insert(self.root, node)
		# balance the tree to avoid degeneration
		self._rotate()

	def find(self, *args):
		"""
		Find all intervals in the tree that intersect with the given interval. *args can be a start and endpoint given as ints, or a single object with the attributes start and end.
		"""
		for interval in self._find(self.root, self._convert_args(*args)):
			yield interval

	def _insert(self, root, node):
		if not root:
			self.root = node
			return
		maxend = None
		if node.interval.start < root.interval.start:
			if root.left:
				maxend = self._insert(root.left, node)
			else:
				root.left = node
		else:
			if root.right:
				maxend = self._insert(root.right, node)
			else:
				root.right = node

		root.height += 1		
		
		if not maxend:
			maxend = node.interval.end

		root.maxend = max(maxend, root.maxend)
		return root.maxend

	def _find(self, root, interval):
		if not root:
			return

		if interval.start > root.maxend:
			return

		if root.interval.is_overlap(interval):
			yield root.interval

		if interval.start < root.interval.start:
			for i in self._find(root.left, interval):
				yield i

		elif not interval.end < root.interval.start:
			for i in self._find(root.right, interval):
				yield i

	def _rotate(self):
		if self.root.right and self.root.left:
			heightdiff = self.root.right.height - self.root.left.height
			if heightdiff > 1:
				# perform a left rotation
				p = self.root
				q = self.root.right
				self.root = q
				p.right = q.left
				q.left = p
				q.height += 1
				p.height -= 1

				p.recalc_maxend()
				q.recalc_maxend()
			elif heightdiff < -1:
				# perform a right rotation
				p = self.root
				q = self.root.left
				self.root = q
				p.left = q.right
				q.right = p
				q.height += 1
				p.height -= 1
				
				p.recalc_maxend()
				q.recalc_maxend()

			
			
			
	
	def _inorder(self, root):
		if not root:
			return

		for i in self._inorder(root.left):
			yield i

		yield root

		for i in self._inorder(root.right):
			yield i

	def __iter__(self):
		for i in self._inorder(self.root):
			yield i

	def __repr__(self):
		return ", ".join(map(str, self))
		
	def _convert_args(self, *args, obj = None, node = False):
		""" Convert args into an interval. *args can be a start and endpoint given as ints, or a single object with the attributes start and end. """
		if len(args) == 2:
			interval = Interval(*args)
		else:
			interval =  Interval(args[0].start, args[0].end)
		if node:
			if not obj and len(args) == 1:
				obj = args[0]
			return IntervalNode(interval, obj, self)
		return interval



from setuptools import setup, Extension
from distutils.version import LooseVersion
from glob import glob
import os
import sys
from sqt import __version__

MIN_CYTHON_VERSION = '0.17'

if sys.version_info < (3, 3):
	sys.stdout.write("At least Python 3.3 is required.\n")
	sys.exit(1)


def out_of_date(extensions):
	"""
	Check whether any pyx source is newer than the corresponding generated
	C(++) source or whether any C(++) source is missing.
	"""
	for extension in extensions:
		for pyx in extension.sources:
			path, ext = os.path.splitext(pyx)
			if ext not in ('.pyx', '.py'):
				continue
			csource = path + ('.cpp' if extension.language == 'c++' else '.c')
			if not os.path.exists(csource) or (
				os.path.getmtime(pyx) > os.path.getmtime(csource)):
				return True
	return False


def no_cythonize(extensions, **_ignore):
	"""
	Change file extensions from .pyx to .c or .cpp.

	Copied from Cython documentation
	"""
	for extension in extensions:
		sources = []
		for sfile in extension.sources:
			path, ext = os.path.splitext(sfile)
			if ext in ('.pyx', '.py'):
				if extension.language == 'c++':
					ext = '.cpp'
				else:
					ext = '.c'
				sfile = path + ext
			sources.append(sfile)
		extension.sources[:] = sources
	return extensions


def cythonize_if_necessary(extensions):
	if '--cython' in sys.argv:
		sys.argv.remove('--cython')
	elif out_of_date(extensions):
		sys.stdout.write('At least one C source file is missing or out of date.\n')
	else:
		return no_cythonize(extensions)

	try:
		from Cython import __version__ as cyversion
	except ImportError:
		sys.stdout.write(
			"ERROR: Cython is not installed. Install at least Cython version " +
			str(MIN_CYTHON_VERSION) + " to continue.\n")
		sys.exit(1)
	if LooseVersion(cyversion) < LooseVersion(MIN_CYTHON_VERSION):
		sys.stdout.write(
			"ERROR: Your Cython is at version '" + str(cyversion) +
			"', but at least version " + str(MIN_CYTHON_VERSION) + " is required.\n")
		sys.exit(1)

	from Cython.Build import cythonize
	return cythonize(extensions)


extensions = [
	Extension('sqt._helpers', sources=['sqt/_helpers.pyx']),
	Extension('sqt._qualtrim', sources=['sqt/_qualtrim.pyx']),
]
extensions = cythonize_if_necessary(extensions)


setup(
	name = 'sqt',
	version = __version__,
	author = 'Marcel Martin',
	author_email = 'marcel.martin@scilifelab.se',
	url = 'https://bitbucket.org/marcelm/sqt',
	description = 'Command-line tools for the analysis of high-throughput sequencing data',
	license = 'MIT',
	packages = [ 'sqt', 'sqt.io', 'sqt.scripts' ],
	scripts = [ s for s in glob(os.path.join("bin", "sqt-*")) if not s.endswith('~') ],
	ext_modules = extensions,
	test_suite = 'nose.collector',
	classifiers = [
		"Development Status :: 3 - Alpha",
		#Development Status :: 4 - Beta
		#Development Status :: 5 - Production/Stable
		"Environment :: Console",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Natural Language :: English",
		"Programming Language :: Python :: 3",
		"Topic :: Scientific/Engineering :: Bio-Informatics"
	]
)



#!/usr/bin/env python3
"""
Compare two strings using global or semi-global alignment.
"""
from sqt import HelpfulArgumentParser
import sys
import os
from sqt.align import globalalign, GlobalAlignment
#, print_alignment #, SEMIGLOBAL, GLOBAL
from sqt.dna import reverse_complement

__author__ = "Marcel Martin"


def align_and_print(sequence1, sequence2, width, semiglobal):
	ga = GlobalAlignment(sequence1, sequence2, semiglobal=semiglobal)
	ga.print(width=width, gap_char='-')
	print()
	print('Length of sequence 1 (top):', len(sequence1))
	print('Length of sequence 2 (bottom):', len(sequence2))
	print('Errors in alignment:', ga.errors)
	print('Length of overlap in sequence 1:', ga.stop1 - ga.start1)
	print('Length of overlap in sequence 2:', ga.stop2 - ga.start2)
	#print('insert size:', front_length, '+', middle_length, '+', tail_length, '=', front_length + middle_length + tail_length)


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument('--semiglobal', action='store_true', default=False,
		help='Run a semi-global alignment (for detecting overlaps)')
	parser.add_argument('--reverse-complement', '--rc', action='store_true', default=False,
		help='Run the alignment also with the reverse-complement of the second sequence')
	parser.add_argument('sequence1')
	parser.add_argument('sequence2')
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()

	sequence1 = args.sequence1.encode('ascii')
	sequence2 = args.sequence2.encode('ascii')

	# credit: http://stackoverflow.com/questions/566746/
	rows, columns = os.popen('stty size', 'r').read().split()
	columns = int(columns)

	print('Alignment:')
	print()
	align_and_print(sequence1, sequence2, columns, semiglobal=args.semiglobal)

	if args.reverse_complement:
		sequence2_rc = reverse_complement(sequence2)
		print()
		print("Alignment with reverse-complemented sequence 2:")
		print()
		align_and_print(sequence1, sequence2_rc, columns, semiglobal=args.semiglobal)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Read a SAM file from standard input, replace all characters in all reads
that are not one of {a, c, g, t, n, A, C, G, T, N} with the character 'N'.
Write the modified SAM file to standard output.

This is approx. 8 times faster than an equivalent awk line using the gsub()
function.
"""
import sys
from os.path import join, dirname, realpath, isfile

from sqt import HelpfulArgumentParser


def main():
	parser = HelpfulArgumentParser(usage=__doc__)
	args = parser.parse_args()

	tab = [ord('N')] * 256
	for c in b'ACGTNacgtn':
		tab[c] = c
	trans = bytes(tab)

	for line in sys.stdin.buffer:
		if line.startswith(b'@'):
			sys.stdout.buffer.write(line)
		else:
			fields = line.split(b'\t')
			fields[9] = fields[9].translate(trans)
			sys.stdout.buffer.write(b'\t'.join(fields))


if __name__ == '__main__':
	main()



"""
Colorspace conversion routines.

Inspired by agapython/util/Dibase.py from Corona lite,
but reimplemented to avoid licensing issues.

Encoding Table

  A C G T
A 0 1 2 3
C 1 0 3 2
G 2 3 0 1
T 3 2 1 0
"""

import string
import sys

__author__ = 'Marcel Martin'

if sys.version > '3':
	xrange = range


def _initialize_dicts():
	"""
	Create the colorspace encoding and decoding dictionaries.
	"""
	enc = {}
	for i, c1 in enumerate("ACGT"):
		enc['N' + c1] = '4'
		enc[c1 + 'N'] = '4'
		enc['.' + c1] = '4'
		enc[c1 + '.'] = '4'
		for j, c2 in enumerate("ACGT"):
			# XOR of nucleotides gives color
			enc[c1 + c2] = chr(ord('0') + (i ^ j))
	enc.update({ 'NN': '4', 'N.': '4', '.N': '4', '..': '4'})

	dec = {}
	for i, c1 in enumerate("ACGT"):
		dec['.' + str(i)] = 'N'
		dec['N' + str(i)] = 'N'
		dec[c1 + '4'] = 'N'
		for j, c2 in enumerate("ACGT"):
			# XOR of nucleotides gives color
			dec[c1 + chr(ord('0') + (i ^ j))] = c2
	dec['N4'] = 'N'

	return (enc, dec)


def encode(s):
	"""
	Given a sequence of nucleotides, convert them to
	color space. Only uppercase characters are allowed.
	>>> encode("ACGGTC")
	"A13012"
	"""
	if not s:
		return s
	r = s[0]
	for i in range(len(s) - 1):
		r += ENCODE[s[i:i+2]]
	return r


def decode(s):
	"""
	Decode a sequence of colors to nucleotide space.
	The first character in s must be a nucleotide.
	Only uppercase characters are allowed.
	>>> decode("A13012")
	"ACGGTC"
	"""
	if len(s) < 2:
		return s
	x = s[0]
	result = x
	for c in s[1:]:
		x = DECODE[x + c]
		result += x
	return result


(ENCODE, DECODE) = _initialize_dicts()

if sys.version > '3':
	# convert to "bytes"
	def _str_dict_to_bytes(d):
		return dict((bytes(k, 'ascii'), bytes(v, 'ascii')) for k,v in d.items())
	ENCODE.update(_str_dict_to_bytes(ENCODE))
	DECODE.update(_str_dict_to_bytes(DECODE))
